#!/usr/bin/perl
################
## wordloader.cgi: 単語アップローダー本体 CGI
## Version 2010.1.8
## (C) 2009 MGJ Interactive
## <http://www.mgjshop.co.jp/mgji/>
################
## バージョン履歴
## 2009.5.15:	初期リリース
## 2009.5.16:	iPhone OS の不具合に対処するための escape サブルーチンの変更
## 2010.1.8:	UTF-8 形式ファイルに対応するための utf16toutf8 サブルーチンの変更
################

################
#### 環境設定
################
## 動作環境のディレクトリー構成に合わせて以下の定義文（ならびに冒頭の perl パス）を適宜書き換えること
# （ディレクトリーパスは絶対パスで指定し末尾の / は省略する）
# （本スクリプトと同じディレクトリーに設置する場合は「.」を指定する）
$DIRSBIN = '/usr/sbin';	# システムバイナリーディレクトリー（sendmail の設置場所）
$DIRLIB = '.';		# ライブラリーディレクトリー（cgi-lib.pl の設置場所）
$DIRTMPL = '.';		# テンプレートディレクトリー（wordloader-tmpl.html の設置場所）
## メール送信処理の送信元メールアドレスを以下に設定すること
# （送信元メールアドレスを正しく設定しないと，不正メール送信とみなされて送信処理が拒否されたりサーバー利用停止となる場合があるので，共用サーバーを利用する場合は特に注意すること）
$FROMADDRESS = 'xxx@xxx.ne.jp';
################

################
#### スクリプト本体
################

## 諸定数の定義
# スクリプト記法上の言語固有リテラル値
$TRUE = 1;			# 真値リテラル
$FALSE = 0;			# 偽値リテラル
$NULL = 0;			# 空値リテラル
$FIRST_INDEX = 0;		# 配列番号の始値
$FIRST_COUNT = 1;		# 一般序数の始値
$COUNT_OFFSET = $FIRST_COUNT - $FIRST_INDEX;
				# 配列番号に対する一般序数のオフセット値
# 外部要素に依存する設定値
$CHECKED = '1';			# フォームのチェックボックスが返す真値
$HTMLTMPL = 'wordloader';	# HTML テンプレート名
$MAILWIDTH = 76;		# メールの折り返し行幅
# メール送信処理で使用するヘッダー情報
$MAILSUBJECT = 'MGJ Interactive: Wordloader Bookmark';
$MAILER = 'MGJ Interactive Wordloader 2010.1.8';
# 機種・ブラウザー判別フラグ
$OTHER = 0;			# その他機種・ブラウザー
$IPHONE = -1;			# iPhone/iPhone 用 Safari
$MAC = 1;			# Mac
$WINDOWS = 2;			# Windows
$SAFARI = 1;			# パソコン用 Safari
$IE = 2;			# Internet Explorer

## ライブラリーの読込み
require "$DIRLIB/cgi-lib.pl";

## HTTP ヘッダーの出力
print "Content-type: text/html\n";
print "Pragma: no-cache\n";
print "Cache-Control: no-store, no-cache\n";
print "\n";

## 使用機種の判定
$agent = $ENV{'HTTP_USER_AGENT'};
if ($agent =~ /\WiPhone\W|\WiPod\W/) { $platform = $IPHONE; }
elsif ($agent =~ /\WMac[^a-z]|\WMacintosh\W/) { $platform = $MAC; }
elsif ($agent =~ /\WWin/) { $platform = $WINDOWS; }
else { $platform = $OTHER; }
## 使用ブラウザーの判定（iPhone 用 Safari はパソコン用とは別区分とする）
if ($agent =~ /\WiPhone\W|\WiPod\W/) { $browser = $IPHONE;}
elsif ($agent =~ /\WSafari\W/) { $browser = $SAFARI; }
elsif ($agent =~ /\WMSIE\W/) { $browser = $IE; }
else { $browser = $OTHER; }

# パラメーターの読込み
&ReadParse;
# 単語リストの取得
&getwordlist;
# メール送信ボタンが押されている場合はメール送信処理を実行し結果フラグを取得する
$mailresult = $FALSE;
if (exists ($in{'send'})) { $mailresult = &sendmail; }

#### HTML テンプレートを加工して表示する
&showhtml ($HTMLTMPL);

exit;

################
#### 以下サブルーチン定義
################

#### HTML テンプレートの加工・出力処理
sub showhtml {
	## 引数の取得
	local $tmplname = shift;	# サブルーチンの引数: テンプレート名
					# （必ず実在する HTML ファイル名を指定すること）

	# ローカル変数の宣言
	local $currentblock = '';
	local $negative;
	local $blockname;
	local $tmpltext = '';

	open(TEMPLATE, "$DIRTMPL/$tmplname.html") || die 'File Error';
	# テンプレートファイルから 1 行づつ読み込んで処理を行う
	while(<TEMPLATE>)
	{
		# if ブロック開始タグを読み込んだ場合
		if( /^[\t ]*<!--[\t ]*#if[\t ]+([Nn][Oo][Tt][\t ]+)?(.*)[\t ]*-->/ ) {
			# ブロック指示子を設定する
			$currentblock = 'if';
			# NOT 指定の場合は否定指示子を設定する
			$negative = length ($1) > 0;
			# ブロック名を取得する
			$blockname = $2;
			next;
		}
		# if ブロック終了タグを読み込んだ場合
		elsif( /^[\t ]*<!--[\t ]*#endif[\t ]+.*-->/ ) {
			# ブロック指示子をリセットする
			$currentblock = '';
			next;
		}
		# template ブロック開始タグを読み込んだ場合
		elsif( /^[\t ]*<!--[\t ]*#template[\t ]+(.*)[\t ]*-->/ ) {
			# ブロック指示子を設定する
			$currentblock = 'template';
			# ブロック名を取得する
			$blockname = $1;
			# テンプレートをリセットする
			$tmpltext = '';
			next;
		}
		# template ブロック終了タグを読み込んだ場合
		elsif( /^[\t ]*<!--[\t ]*#endtemplate[\t ]+.*-->/ ) {
			# ブロック名によって後処理を振り分ける
			# 単語入力欄の場合は入力欄のテーブル行を構築する
			if ($blockname eq 'wordtable') { &createwordtable ($tmpltext); }			
			# ブックマークの場合はブックマークアイコンを表示する（ブックマーク作成済の場合）
			elsif ($blockname eq 'bookmark' && exists ($in{'bookmark'}) && scalar (@wordlist) > 0) {
				printf $tmpltext, &escapedwordlist; 
			}
			# メール送信欄の場合はメールアドレスを再設定する
			elsif ($blockname eq 'sendmail') {
				printf $tmpltext, &htmlescape ($in{'email'});
			}
			# デバッグ情報の場合は指定したテキストを埋め込む
			elsif ($blockname eq 'debug') {
				# デバッグ情報を表示する場合は下文のコメントを解除し第 2 引数にデバッグ情報を指定する 
				# printf $tmpltext, "Agent = $agent";
				# printf $tmpltext, "Platform = $platform, Browser = $browser";
			}
			# ブロック指示子をリセットする
			$currentblock = '';
			next;
		}

		## 現在処理中のブロック指示子によって処理を振り分ける
		# if ブロックの場合
		if ($currentblock eq 'if') {
			# セキュアーページ専用ブロックに関する処理
			if ($blockname eq 'secure') {
				# 表示中ページが https でない場合はセキュアーページ専用ブロックをスキップする
				if (!$negative && $ENV{'HTTPS'} ne 'on') { next; }
				# 表示中ページが https の場合は非セキュアーページ専用ブロックをスキップする
				if ($negative && $ENV{'HTTPS'} eq 'on') { next; }
			}
			# ブックマーク作成ブロックに関する処理
			elsif ($blockname eq 'bookmark') {
				# ブックマーク未作成場合はブックマーク作成済ブロックをスキップする
				if (!$negative && !(exists ($in{'bookmark'})  && scalar (@wordlist) > 0)) { next; }
				# ブックマーク作成済場合はブックマーク未作成ブロックをスキップする
				if ($negative && exists ($in{'bookmark'})  && scalar (@wordlist) > 0) { next; }
			}
			# メール送信ブロックに関する処理
			elsif ($blockname eq 'mailsent') {
				# メール未送信の場合はメール送信ブロックをスキップする
				if (!$negative && !(exists ($in{'send'}) && $mailresult)) { next; }
				# メール送信済の場合はメール未送信ブロックをスキップする
				if ($negative && exists ($in{'send'}) && $mailresult) { next; }
			}
			# メール送信失敗ブロックに関する処理
			elsif ($blockname eq 'mailfailure') {
				# メール未送信または送信成功の場合はメール送信失敗ブロックをスキップする
				if (!$negative && !(exists ($in{'send'}) && !$mailresult)) { next; }
				# メール送信を失敗した場合はメール送信無失敗ブロックをスキップする
				if ($negative && exists ($in{'send'}) && !$mailresult) { next; }
			}
			# iPhone 用ブロックに関する処理
			elsif ($blockname eq 'platform iphone') {
				# 使用機種が iPhone でない場合は iPhone 用ブロックをスキップする
				if (!$negative && $platform != $IPHONE) { next; }
			}
			# Mac 用ブロックに関する処理
			elsif ($blockname eq 'platform mac') {
				# 使用機種が Mac でない場合は Mac 用ブロックをスキップする
				if (!$negative && $platform != $MAC) { next; }
			}
			# Windows 用ブロックに関する処理
			elsif ($blockname eq 'platform windows') {
				# 使用機種が Windows でない場合は Windows 用ブロックをスキップする
				if (!$negative && $platform != $WINDOWS) { next; }
			}
			# その他の機種用ブロックに関する処理
			elsif ($blockname eq 'platform other') {
				# 使用機種がその他の機種でない場合はその他の機種用ブロックをスキップする
				if (!$negative && $platform != $OTHER) { next; }
			}
			# Mac 版その他ブラウザー用ブロックに関する処理
			elsif ($blockname eq 'browser mac other') {
				# 使用ブラウザーが Mac 版の Safiri 以外のブラウザーでない場合はその他　Mac 版その他ブラウザー用ブロックをスキップする
				if (!$negative && !($platform == $MAC && $browser != $SAFARI)) { next; }
			}
			# Windows 版その他ブラウザー用ブロックに関する処理
			elsif ($blockname eq 'browser windows other') {
				# 使用ブラウザーが Windows 版の Sarari あるいは Internet Explorer の場合は Windows 版その他ブラウザー用ブロックをスキップする
				if (!$negative && !($platform == $WINDOWS && $browser == $OTHER)) { next; }
			}
			# Safari 用ブックマーク作成ブロックに関する処理
			elsif ($blockname eq 'bookmark safari') {
				# ブックマークが作成済かつ使用ブラウザーが Safari でない場合は Safari 用ブロックをスキップする
				if (!$negative && !(exists ($in{'bookmark'}) && scalar (@wordlist) > 0 && $browser == $SAFARI)) { next; }
			}
			# Internet Explorer 用ブックマーク作成ブロックに関する処理（Mac 版 Internet Explorer は除外する）
			elsif ($blockname eq 'bookmark ie') {
				# ブックマークが作成済かつ使用ブラウザーが Windows 版 Internet Explorer でない場合は Internet Explorer 用ブロックをスキップする
				if (!$negative && !(exists ($in{'bookmark'}) && scalar (@wordlist) > 0 && $platform == $WINDOWS && $browser == $IE)) { next; }
			}
			# Mac 版その他ブラウザー用ブックマーク作成ブロックに関する処理
			elsif ($blockname eq 'bookmark mac other') {
				# ブックマークが作成済かつ使用ブラウザーが Mac 版の Safiri 以外のブラウザーでない場合はその他ブラウザー用ブックマーク作成ブロックをスキップする
				if (!$negative && !(exists ($in{'bookmark'}) && scalar (@wordlist) > 0 && $platform == $MAC && $browser != $SAFARI)) { next; }
			}
			# Windows 版その他ブラウザー用ブックマーク作成ブロックに関する処理
			elsif ($blockname eq 'bookmark windows other') {
				# ブックマークが作成済かつ使用ブラウザーが Windows 版の Sarari あるいは Internet Explorer の場合 Windows 版その他ブラウザー用ブロックをスキップする
				if (!$negative && !(exists ($in{'bookmark'}) && scalar (@wordlist) > 0 && $platform == $WINDOWS && $browser == $OTHER)) { next; }
			}
			# iPhone 用ブックマーク作成ブロックに関する処理
			elsif ($blockname eq 'bookmark iphone') {
				# ブックマークが作成済かつ使用機種が iPhone でない場合は iPhone 用ブックマーク作成ブロックをスキップする
				if (!$negative && !(exists ($in{'bookmark'}) && scalar (@wordlist) > 0 && $platform == $IPHONE)) { next; }
			}
			# その他機種用ブックマーク作成ブロックに関する処理
			elsif ($blockname eq 'bookmark other') {
				# ブックマークが作成済かつ使用機種が Mac または Windows の場合はその他機種用ブックマーク作成ブロックをスキップする
				if (!$negative && !(exists ($in{'bookmark'}) && scalar (@wordlist) > 0 && $platform == $OTHER)) { next; }
			}
			# それ以外のブロック名については全てスキップする
			else { next; }
		}
		# template ブロックの場合
		elsif ($currentblock eq 'template') {
			# 現在の行をテンプレートに追加する
			$tmpltext .= $_;
			next;
		}

		# 行が置換文字列を含まない場合はテンプレートをそのまま出力する
		print $_;
	}
	close(TEMPLATE);
}

## パラメーター入から単語リストを取得して，グローバル変数 @wordlist に設定する処理
sub getwordlist {
	## 本サブルーチンでは以下のグローバル変数を使用する
	# @wordlist: パラメーターから取得する入力単語リスト
	# $rowcount: 単語入力欄の件数

	## 定数の定義
	local $INITIALROWS = 8; # 入力欄の初期件数

	# グローバル変数の初期化
	@wordlist = ();
	$rowcount = 0;

	local $reading;
	local $word;
	local $check;
	local $delete = exists ($in{'delete'});
	local $lastentry = 0;
	local $rownumber;
	for ($rownumber = $FIRST_COUNT; defined ($in{'r'.$rownumber}); ++$rownumber) {
		$reading = $in{'r'.$rownumber};
		$word = $in{'w'.$rownumber};
		$check = $in{'c'.$rownumber};
		if (!$delete || $check ne $CHECKED) {
			++$rowcount;
			push (@wordlist, "$reading\t$word\t$check");
			if (length ($reading) > 0 || length ($word) > 0) { $lastentry = $rowcount; }
		}
	}
	if ($rowcount < $INITIALROWS) { $rowcount = $INITIALROWS }
	splice (@wordlist, $lastentry);
}

## メール送信処理
sub sendmail {
	## メールアドレスの有効性チェック
	local $email = $in{'email'};
	# 前後の空白は除去する
	$email =~ s/^\s*(\S*)\s*$/$1/;
	# メールアドレスが RFC2822 に準拠していない場合は以後の処理は実行せず否定値を返す
	if ($email !~/^[\w!#\$%&'*+\-\/=?^`{|}~]([\w!#\$%&'*+\-\/=?^`{|}~.]*[\w!#\$%&'*+\-\/=?^`{|}~])?@[\w!#\$%&'*+\-\/=?^`{|}~]([\w!#\$%&'*+\-\/=?^`{|}~.]*[\w!#\$%&'*+\-\/=?^`{|}~])?$/) {
		return $FALSE;
	}

	## メール送信処理
	open(MAIL,"| $DIRSBIN/sendmail -t") || die 'Sendmail Error';
# メールヘッダーを出力
# ================
print MAIL <<"EOF";
Subject: $MAILSUBJECT
From: $FROMADDRESS
To: $email
X-Mailer: $MAILER
Content-Type: text/html; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit

EOF
# ================

	# URL エスケープした単語リストを取得する
	local $escapedstring = &escapedwordlist;
	# メールの折り返し行幅ごとに改行を挿入する
	$escapedstring =~ s/.{$MAILWIDTH}/$&\n/g;

# メール本文を出力
# ================
print MAIL <<"EOF";
<HTML>
	<BODY>
		<P><A href="mgjiap://jp.co.mgjshop.WordRegisterPlus?action=gettext&amp;data=
$escapedstring
">&#21336;&#35486;&#12487;&#12540;&#12479;</A></P>
	</BODY>
</HTML>
EOF
# ================

	close(MAIL);

	return $TRUE;
}

## 単語リストから URL エスケープした文字列を構築する処理
sub escapedwordlist {
	## 本サブルーチンでは以下のグローバル変数を使用する
	# @wordlist: パラメーターから取得した入力単語リスト

	## 単語リストの各要素について処理する
	local $listtext = '';
	local $wordrecord;
	foreach $wordrecord (@wordlist) {
		# 各要素のタブで区切られた最初の 2 項目を抽出し，改行を付加してテキストに追加する
		$wordrecord =~ s/(^[^\t]*\t[^\t]*).*/$1/;
		$listtext .= "$wordrecord\n";
	}

	# URL エスケープしたテキストを返す
	return &escape ($listtext);
}

## 単語入力欄のテーブル行を構築する処理
sub createwordtable {
	## 定数の定義
	local $SELECTED = 1;
	local $DESELECTED = -1;

	## 本サブルーチンでは以下のグローバル変数を使用する
	# @wordlist: パラメーターから取得した入力単語リスト
	# $rowcount: 単語入力欄の件数

	## 引数の取得
	local $tmpltext = shift;

	local $selectall = $NULL;
	if (exists ($in{'append'})) {
		local $rows = $in{'rows'};
		if ( 0 < $rows && $rows < 100 ) { $rowcount += $rows; }
	}
	elsif (exists ($in{'trim'})) {
		$rowcount = $INITIALROWS;
	}
	elsif (exists ($in{'import'})) {
		local $utf8 = &utf16toutf8 ($in{'file'});
		push (@wordlist, &parsewordslist ($utf8));
	}
	elsif (exists ($in{'selectall'})) {
		$selectall = $SELECTED;
	}
	elsif (exists ($in{'deselectall'})) {
		$selectall = $DESELECTED;
	}

	if ($rowcount <= scalar(@wordlist)) { $rowcount = scalar(@wordlist) + 1; }

	local $checkstatus;
	local $wordnumber;
	for ($wordnumber = $FIRST_COUNT; $wordnumber <= $rowcount; ++$wordnumber) {
		($reading, $word, $check) = split (/\t/, $wordlist[$wordnumber - $COUNT_OFFSET]);
		if ($check eq $CHECKED && $selectall !=  $DESELECTED || $selectall == $SELECTED) { $checkstatus = ' checked'; }
		else { $checkstatus = ''; }
		printf $tmpltext, $wordnumber, $checkstatus, $wordnumber, &htmlescape ($reading), $wordnumber, &htmlescape ($word);
	}
}

#### UTF-16 を UTF-8 に変換する関数
sub utf16toutf8 {
	## 定数の定義
	local $BE_BOM = 0xFEFF;	# BOM 文字コード（正順形式の BOM コード）
	local $LE_BOM = 0xFFFE;	# 逆 BOM 文字コード（逆順形式の BOM コード）
	local $U_START_2 = 0x80;	# UTF-8 2 バイト文字領域の開始位置
	local $U_START_3 = 0x800;	# UTF-8 3 バイト文字領域の開始位置
	local $U_START_SUR = 0xD800;	# UTF-16 代用文字領域の開始位置
	local $U_END_SUR = 0xE000;	# UTF-16 代用文字領域の終了位置の次位
	local $U_START_LSUR = 0xDC00;	# UTF-16 低位代用文字領域の開始位置
	local $F_HEADER_2 = 0xC0;	# UTF-8 ヘッダーバイト識別フラグ（2 バイト文字用）
	local $F_HEADER_3 = 0xE0;	# UTF-8 ヘッダーバイト識別フラグ（3 バイト文字用）
	local $F_HEADER_4 = 0xF0;	# UTF-8 ヘッダーバイト識別フラグ（4 バイト文字用）
	local $F_TRAILER = 0x80;	# UTF-8 トレーラーバイト識別フラグ
	local $M_2BITS = 0x3; 	# 下位 2 ビットマスク
	local $M_4BITS = 0xF;	# 下位 4 ビットマスク
	local $M_6BITS = 0x3F;	# 下位 6 ビットマスク

	## 引数の取得と返値の初期化
	local $utf16 = shift;	# 関数の引数: UTF-16 文字列
	local $utf8 = '';	# 関数の返値: UTF-8 文字列

	## ファイル先頭ワードの BOM コードを調べて，文字コード列を配列に格納する処理
	local $isUTF16 = $TRUE;
	local @wordcodes;
	# ファイル先頭ワードを正順形式で取得する
	local @bom = unpack ("n", $utf16);
	# 正順形式の BOM コードに一致する場合はファイル全体を正順形式で読み込む
	if ($bom[$FIRST_INDEX] == $BE_BOM) {
		@wordcodes = unpack ("n*", $utf16);
		shift (@wordcodes);
	}
	# 逆順形式の BOM コードに一致する場合はファイル全体を逆順形式で読み込む
	elsif ($bom[$FIRST_INDEX] == $LE_BOM) {
		@wordcodes = unpack ("v*", $utf16);
		shift (@wordcodes);
	}
	# 先頭文字が BOM でない場合ファイル形式を UTF-8 と仮定し変換せずにそのまま返す
	else {
		$isUTF16 = $FALSE;
		$utf8 = $utf16;
	}

	if ($isUTF16) {
		## 文字コード列の各要素を調べて UTF-8 バイト列を構築する処理
		local $plane;
		local $hsurcode = $NULL;
		local $wordcode;
		# 文字コード列の各要素について処理を行う
		foreach $wordcode (@wordcodes) {
			# 1 バイト文字の場合は文字コードをそのまま追加する
			if ($wordcode < $U_START_2) {
				$hsurcode = $NULL;
				$utf8 .= chr ($wordcode);
			}
			# 2 バイト文字の場合は下位 6 ビットとその上位に分割し，
			# それぞれ識別フラグを付けた文字コードを追加する
			elsif ($wordcode < $U_START_3) {
				$hsurcode = $NULL;
				$utf8 .= chr ($F_HEADER_2 | $wordcode >> 6);
				$utf8 .= chr ($F_TRAILER | $wordcode & $M_6BITS);
			}
			# 3 バイト文字の場合は下位 6 ビットと中位 6 ビットとその上位に分割し，
			# それぞれ識別フラグを付けた文字コードを追加する
			elsif ($wordcode < $U_START_SUR || $U_END_SUR <= $wordcode) {
				$hsurcode = $NULL;
				$utf8 .= chr ($F_HEADER_3 | $wordcode >> 12);
				$utf8 .= chr ($F_TRAILER | $wordcode >> 6 & $M_6BITS);
				$utf8 .= chr ($F_TRAILER | $wordcode & $M_6BITS);
			}
			# 高位代用文字の場合は処理をスキップして次の文字コード処理に引き渡す
			elsif ($wordcode < $U_START_LSUR) {
				$hsurcode = $wordcode;
			}
			# 低位代用文字の場合で直前の文字が高位代用文字の場合
			elsif ($NULL < $hsurcode) {
				# 高位代用文字の 7 ビット目から 4 ビット分に 1 を加え拡張領域の面番号とする
				$plane = $COUNT_OFFSET + ($hsurcode >> 6 & $M_4BITS);
				# 面番号の上位 3 ビット目以上，面番号の下位 2 ビット，高位代用文字の 3 ビット目から 4 ビット分，
				# 高位代用文字の下位 2 ビット，低位代用文字の 7 ビット目から 4 ビット分，
				# 低位代用文字の下位 6 ビットを，それぞれ 4 バイト分に組み合わせて識別フラグを付ける
				$utf8 .= chr ($F_HEADER_4 | $plane >> 2);
				$utf8 .= chr ($F_TRAILER | ($plane & $M_2BITS) << 4 | $hsurcode >> 2 & $M_4BITS);
				$utf8 .= chr ($F_TRAILER | ($hsurcode & $M_2BITS) << 4 | $wordcode >> 6 & $M_4BITS);
				$utf8 .= chr ($F_TRAILER | $wordcode & $M_6BITS);
				$hsurcode = $NULL;
			}
		}
	}
	# 構築した UTF-8 文字列を返す
	return $utf8;
}

#### 単語一覧データを，よみと登録語からなるタブ区切りリストに規格化する関数
# （データのエンコーディングは UTF-8 を前提とし，CSV の場合はフィールドにタブ制御文字を含まないこと）
sub parsewordslist {
	## 定数の定義
	local $HT = "\t";		# HT 制御文字（"\x09"）
	local $LF = "\n";		# LF 制御文字（"\x0A"）
	local $CR = "\r";		# CR 制御文字（"\x0D"）
	local $QQ = '"';		# 引用符文字
	local $CM = ',';		# カンマ文字
	local $PILCROW = "\xC2\xB6";	# 改行置換文字（¶）の UTF-8 表記

	## 引数の取得と返値の初期化
	local @lines = split (/[$LF$CR]|$CR$LF/, shift);	# 関数の引数: ファイルから取得した単語一覧データ
								# （あらかじめ行ごとに配列として格納しておく）	
	local @resultdata = ();					# 関数の返値: よみと登録語からなるタブ区切りデータのリスト

	## 引数で渡された各行データに対する処理
	local $line;
	local $record = '';
	local $reading;
	local $word;
	local $dummy;
	foreach $line (@lines) {
		# 行データがタブ制御文字を含む場合はタブ区切りテキストとして最初の 2 フィールドを取得する
		if (split (/$HT/, $line) >= 2) {
			$record = '';
			($reading, $word) = split (/$HT/, $line);
		}
		# 行データがタブ文字を含まない場合は CSV として処理を試みる
		else {
			# レコードに行データを追加する
			$record .= $line;
			# レコード内の引用句が完結している場合（全ての引用符が対になっている場合）は，
			# 完成したレコードとして CSV 分解処理を実行する
			if (($record =~ tr/"/"/) % 2 == 0) {
				# 次のパターン検索のためにレコードの末尾に区切記号（カンマ）を追加する
				$record .= $CM;
				# レコードから最初の 2 フィールド（引用句または非引用句）を抽出する
				# （第 2 後方参照（$2）には ($QQ$QQ[^$QQ]*) の内容が設定されているためダミー変数へ破棄する）
				($reading, $dummy, $word) = ($record =~ /($QQ[^$QQ]*($QQ$QQ[^$QQ]*)*$QQ|[^$CM]*)$CM/g);
				# 各フィールドが引用句からなる場合は引用符の内部を抽出する
				$reading =~ s/^$QQ(.*)$QQ$/$1/;
				$word =~ s/^$QQ(.*)$QQ$/$1/;
				# 各フィールド内の連続する引用符は単独の引用符に変換する
				$reading =~ s/$QQ$QQ/$QQ/g;
				$word =~ s/$QQ$QQ/$QQ/g;
				# 処理が完了したレコードの内容はリセットする
				$record = '';
			}
			# レコード内の引用句が完結していない場合（対になっていない引用符が存在する場合）は次の行データに続ける
			else {
				# 結果データ出力をスキップするため，各フィールドには空文字列を設定する
				($reading, $word) = ('', '');
				# 次の行データとの間には改行置換文字を挿入する
				$record .= $PILCROW;
			}
		}
		# 両フィールドが空でない場合は，レコードを結果データとして出力する
		if (0 < length ($reading) && 0 < length ($word)) {
			push (@resultdata, "$reading$HT$word");
		}
	}

	# 結果データを返す
	return @resultdata;
}

#### HTML 不適合文字を実体参照に変換する関数
sub htmlescape {
	$text = shift;

	$text =~ s/&/&amp;/g;
	$text =~ s/"/&quot;/g;
	$text =~ s/</&lt;/g;
	$text =~ s/>/&gt;/g;

	return $text;
}

#### 文字列の各バイトを % エスケープする関数
# （本来この関数は URL 不適合バイトのみエスケープすることを目的としていたが，iPhone OS で特定条件のブックマークが開けないという不具合が判明したため，全バイトをエスケープする処理に変更した。）
# （一般の処理で本関数を利用する場合は，コメントアウトした「オリジナル処理」のブロックを復元し，その下の変更後処理のブロックを削除すること。）
sub escape {
	## 引数の取得と返値の宣言
	local $text = shift;	# 関数の引数ならびに返値: 加工対象のデータ

	################
	#### iPhone OS バグ対策のための処理変更
	################
	#### 文字列中の URL 不適合バイトを % エスケープする処理（オリジナル処理）
	# # スペース以外の URL 不適合バイトを 16 進コード化し先頭に % を付与する 
	# $text =~ s/([^\w.\- ])/'%'.sprintf ('%02X', unpack ('C', $1))/eg;
	# # スペースを「+」で置換する 
	# $text =~ tr/ /+/;
	################
	#### 文字列の全バイトを % エスケープする処理（変更後処理）
	$text =~ s/[\x00-\xFF]/'%'.sprintf ('%02X', unpack ('C', $&))/eg;
	################

	# 加工済データを返す
	return $text;
}
